import email.utils
import uuid
from collections import defaultdict
from datetime import datetime, timezone
from textwrap import dedent
from uuid import UUID

from django.conf import settings
from django.core.exceptions import ValidationError as DjangoValidationError
from django.db import connection, transaction
from django.db.models import CharField, Count, FloatField, Max, Prefetch, Q, QuerySet, prefetch_related_objects
from django.db.models.functions import Cast
from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property
from drf_spectacular.utils import (
    OpenApiParameter,
    PolymorphicProxySerializer,
    extend_schema,
    extend_schema_view,
    inline_serializer,
)
from psycopg2.extras import execute_values
from rest_framework import permissions, serializers, status
from rest_framework.exceptions import NotFound, PermissionDenied, ValidationError
from rest_framework.generics import (
    CreateAPIView,
    DestroyAPIView,
    ListAPIView,
    ListCreateAPIView,
    RetrieveUpdateDestroyAPIView,
    UpdateAPIView,
)
from rest_framework.mixins import DestroyModelMixin
from rest_framework.response import Response

from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.models import (
    AllowedMetaData,
    Classification,
    ClassificationState,
    Corpus,
    Element,
    ElementPath,
    ElementType,
    MetaData,
    MetaType,
    Selection,
    Transcription,
)
from arkindex.documents.serializers.elements import (
    CorpusSerializer,
    ElementBulkSerializer,
    ElementCreateSerializer,
    ElementDestinationSerializer,
    ElementListSerializer,
    ElementNeighborsSerializer,
    ElementParentSerializer,
    ElementSerializer,
    ElementSlimSerializer,
    ElementTypeSerializer,
    MetaDataBulkSerializer,
    MetaDataUpdateSerializer,
)
from arkindex.documents.serializers.light import CorpusAllowedMetaDataSerializer, ElementTypeLightSerializer
from arkindex.documents.serializers.ml import ElementTranscriptionSerializer
from arkindex.images.models import Image
from arkindex.project.mixins import ACLMixin, CorpusACLMixin, SelectionMixin
from arkindex.project.openapi import AutoSchema
from arkindex.project.pagination import LargePageNumberPagination, PageNumberPagination
from arkindex.project.permissions import IsVerified, IsVerifiedOrReadOnly
from arkindex.project.tools import BulkMap
from arkindex.project.triggers import corpus_delete, element_trash, move_element, worker_results_delete
from arkindex.users.models import Role
from arkindex.users.utils import filter_rights

classifications_queryset = Classification.objects.select_related('ml_class', 'worker_version').order_by('-confidence')

best_classifications_prefetch = Prefetch(
    'classifications',
    queryset=classifications_queryset
    # Keep best or validated distinct classifications
    .filter(Q(state=ClassificationState.Validated) | Q(high_confidence=True))
    # Remove rejected classifications
    .exclude(state=ClassificationState.Rejected).distinct(),
    to_attr='best_classes'
)


def _fetch_has_children(elements):
    """
    From a list of elements, sets the `has_children` attribute to a boolean
    indicating if an element has children elements.

    The initial goal was to have a way for the frontend to tell if an element had any children;
    having a count was optional. It is possible to just use an Exists subquery since Django 2.2:
    Element.objects.annotate(has_children=Exists(ElementPath.objects.filter(path__last=OuterRef('pk'))))
    This works but makes the element list queries 20 times slower, due to the way the Postgres query planner
    handles this.

    We used a second query with a GROUP BY to retrieve the element path count, because the difference between
    telling if paths exists and getting a count is negligible with a GROUP BY, but the query could get very slow
    for elements with a lot of children, even without the count.

    We switched back to a boolean and to a query using a LEFT JOIN on a set of values, using the array overlap
    operator (&&) to match any child (not just direct children). This assumes that ElementPath is properly maintained
    and that no "ghost elements" exist in the paths.
    With an array instead of values, Postgres has to resort to a very slow sequential scan over the whole table.
    With direct children (the last element in a path), Postgres cannot use the GIN index on `path` and
    uses a sequential scan and a hash join.

    The ElementsListBase uses arkindex.project.tools.BulkMap to apply this method and perform the second request
    *after* DRF's pagination, because there is no way to perform post-processing after pagination in Django
    without having to use Django private methods.
    """
    # Ignore empty results
    if not elements:
        return elements

    with connection.cursor() as cursor:
        execute_values(
            cursor,
            'SELECT DISTINCT ON (e.id) e.id, p.id is not null as has_children '
            'FROM (VALUES %s) e (id) '
            'LEFT JOIN documents_elementpath p ON (ARRAY[e.id] && p.path)',
            tuple((element.id, ) for element in elements),
        )
        has_children = dict(cursor.fetchall())

    for element in elements:
        element.has_children = has_children[element.id]

    return elements


# Maps valid operator names in the API to Django QuerySet lookups
METADATA_OPERATORS = {
    'eq': 'exact',
    'lt': 'lt',
    'gt': 'gt',
    'lte': 'lte',
    'gte': 'gte',
    # The contains operator should be case-insensitive
    'contains': 'icontains',
}

# Operators that do not require the metadata value to be a number
METADATA_STRING_OPERATORS = {'eq', 'contains'}

ELEMENT_ORDERING_FIELDS = {'name', 'created'}


class ElementsListAutoSchema(AutoSchema):
    """
    AutoSchema subclass that adds common parameters for all element list endpoints.
    This is not doable using @extend_schema as this causes the same schema to be reused for each endpoint
    subclassing ElementsListBase, causing duplicate operation IDs and invalid parameters.
    """

    def get_override_parameters(self):
        # Parameters that apply to all methods
        parameters = [
            OpenApiParameter(
                'type',
                description='Filter elements by type',
                required=False,
            ),
            OpenApiParameter(
                'name',
                description='Only include elements whose name contains a given string (case-insensitive)',
                required=False,
            ),
            OpenApiParameter(
                'folder',
                description='Restrict to or exclude elements with folder types',
                type=bool,
                required=False,
            ),
            OpenApiParameter(
                'worker_version',
                description='Only include elements created by a specific worker version. '
                            'If set to `False`, only include elements created by humans.',
                type=UUID,
                required=False,
            ),
        ]

        # Add method-specific parameters
        if self.method.lower() == 'get':
            parameters.extend([
                OpenApiParameter(
                    'best_class',
                    description='Restrict to or exclude elements with a best class, '
                                'or restrict to elements with specific best class',
                    type={
                        'oneOf': [
                            {'type': 'string', 'format': 'uuid'},
                            {'type': 'boolean'}
                        ]
                    }
                ),
                OpenApiParameter(
                    'metadata_name',
                    description='Restrict to elements having a metadata with the given name.',
                    required=False,
                ),
                OpenApiParameter(
                    'metadata_value',
                    description='Restrict to elements having a metadata with the given value. '
                                'Can be set it exclude elements with a value or filter by numerical values using `metadata_operator`. '
                                'Requires `metadata_name` to be set.',
                    required=False,
                ),
                OpenApiParameter(
                    'metadata_operator',
                    description=dedent("""
                    Set the comparison operator to filter on metadata values:

                    * `eq` (default): Elements having a metadata with the given value.
                    * `contains`: Elements having a metadata whose value contains a given string.
                    * `lt`: Elements having a `numeric` metadata whose value is less than the filter.
                    * `lte`: Elements having a `numeric` metadata whose value is less than or equal to the filter.
                    * `gt`: Elements having a `numeric` metadata whose value is greater than the filter.
                    * `gte`: Elements having a `numeric` metadata whose value is greater than or equal to the filter.

                    This requires `metadata_name` and `metadata_value` to be set.
                    """),
                    enum=METADATA_OPERATORS.keys(),
                    default='eq',
                    required=False,
                ),
                OpenApiParameter(
                    'order',
                    description='Sort elements by a specific field',
                    enum=ELEMENT_ORDERING_FIELDS,
                    default='name',
                    required=False,
                ),
                OpenApiParameter(
                    'order_direction',
                    description='Direction in which to sort elements',
                    enum={'asc', 'desc'},
                    default='asc',
                    required=False,
                ),
                OpenApiParameter(
                    'with_best_classes',
                    description='Returns best classifications for each element. '
                                'If not set, elements best_classes field will always be null',
                    type=bool,
                    required=False,
                ),
                OpenApiParameter(
                    'with_has_children',
                    description='Include the `has_children` boolean to tell if each element has direct children. '
                                'Otherwise, `has_children` will always be null.',
                    type=bool,
                    required=False,
                ),
                OpenApiParameter(
                    'with_corpus',
                    description='Returns the corpus attribute for each element. '
                                'When not set, the corpus attribute will be returned by default.',
                    type=bool,
                    default=True,
                    required=False,
                ),
                OpenApiParameter(
                    'with_zone',
                    description='Returns the zone attribute for each element. '
                                'When not set, the zone attribute will be returned by default.',
                    type=bool,
                    default=True,
                    required=False,
                ),
                OpenApiParameter(
                    'If-Modified-Since',
                    location=OpenApiParameter.HEADER,
                    description='Return HTTP 304 Not Modified when no elements have been created or updated since that time, '
                                'otherwise respond normally. Expects HTTP date formats (`Wed, 21 Oct 2015 07:28:00 GMT`). '
                                'Note that this is unable to detect deletions and will not filter the returned elements.',
                    required=False
                )
            ])

        elif self.method.lower() == 'delete':
            parameters.extend([
                OpenApiParameter(
                    'delete_children',
                    description='Delete all child elements of those elements recursively.',
                    type=bool,
                    required=False,
                )
            ])

        return parameters

    def get_tags(self):
        return ['elements']


class ElementsListBase(CorpusACLMixin, DestroyModelMixin, ListAPIView):
    """
    Provides common features across element list endpoints (list, parents, children)
    """
    schema = ElementsListAutoSchema()
    serializer_class = ElementListSerializer
    queryset = Element.objects.all()
    permission_classes = (IsVerifiedOrReadOnly, )
    pagination_class = LargePageNumberPagination

    def initial(self, *args, **kwargs):
        """
        Adds a `self.clean_params` attribute that holds any query params or headers,
        filtered by their allowed HTTP method.
        Does not perform validation on the filter values.
        """
        super().initial(*args, **kwargs)

        self.clean_params = {}
        # Set attributes for Spectacular's schema generation
        self.schema.path = ''
        self.schema.method = self.request.method
        for param in self.schema._get_parameters():
            if param['in'] == 'header':
                origin_dict = self.request.headers
            elif param['in'] == 'query':
                origin_dict = self.request.query_params
            else:
                raise NotImplementedError

            if param['name'] in origin_dict:
                self.clean_params[param['name']] = origin_dict[param['name']]

    def check_corpus_access(self, corpus):
        """
        Check that the corpus is accessible by the current user.
        This prevents comparing the corpus of every element.
        Require a contributor access for 'unsafe' methods (POST/PUT/PATCH).
        Require an admin access for deletion (DELETE).
        """
        role = Role.Contributor
        if self.request.method in permissions.SAFE_METHODS:
            role = Role.Guest
        elif self.request.method == 'DELETE':
            role = Role.Admin
        if not self.has_access(corpus, role.value):
            raise PermissionDenied(detail=f'You do not have {str(role).lower()} access to this corpus.')

    @cached_property
    def selected_corpus(self):
        # Retrieve the corpus and check user access rights only once
        corpus_id = self.kwargs.get('corpus')
        if corpus_id is None:
            return

        corpus = get_object_or_404(Corpus, id=corpus_id)
        self.check_corpus_access(corpus)
        return corpus

    @property
    def folder_filter(self):
        if self.clean_params.get('folder') is None:
            return
        return self.clean_params['folder'].lower() not in ('false', '0')

    @cached_property
    def type_filter(self):
        if 'type' not in self.request.query_params:
            return
        try:
            element_type = self.selected_corpus.types.get(slug=self.clean_params['type'])
        except ElementType.DoesNotExist:
            raise ValidationError({'type': ['This type does not exist.']})

        # When an API request is made with &type=…&folder=true, we avoid adding a type__folder filder in the query,
        # which would cause a join on ElementType, and instead check for the type's folder state here
        if self.folder_filter is not None and element_type.folder != self.folder_filter:
            raise ValidationError({'non_field_errors': [
                f'This type is {"not " if self.folder_filter else ""} a folder type.']
            })

        return element_type

    def get_worker_version_filter(self):
        # If worker_version query param is False, we have to return elements created by humans
        if self.request.query_params['worker_version'].lower() in ('false', '0'):
            return None

        try:
            worker_version_id = uuid.UUID(self.request.query_params['worker_version'])
        except (TypeError, ValueError):
            raise ValidationError(['Invalid UUID'])

        if not WorkerVersion.objects.filter(id=worker_version_id).exists():
            raise ValidationError(['This worker version does not exist.'])

        return worker_version_id

    def get_metadata_queryset(self):
        """
        Returns a queryset that includes matched element IDs from metadata filters,
        or None if no metadata filters apply.
        """
        name = self.clean_params.get('metadata_name')
        value = self.clean_params.get('metadata_value')
        operator = self.clean_params.get('metadata_operator', '').lower().strip()
        if not any((name, value, operator)):
            return

        queryset = MetaData.objects.all()
        errors = defaultdict(list)

        if name:
            queryset = queryset.filter(name=name)

        if operator:
            if operator not in METADATA_OPERATORS:
                errors['metadata_operator'].append('This metadata operator is not supported.')
            if value is None:
                errors['metadata_operator'].append('This option is not supported without metadata_value.')
            elif operator not in METADATA_STRING_OPERATORS:
                try:
                    value = float(value)
                except (ValueError, TypeError):
                    errors['metadata_value'].append('Value must be a float when using a numeric comparison operator.')
                    value = None
        else:
            operator = 'eq'

        if value:
            lookup = METADATA_OPERATORS[operator]
            if name:
                if operator in METADATA_STRING_OPERATORS:
                    queryset = queryset.filter(**{f'value__{lookup}': value})
                else:
                    # Apply the numeric type filter, and filter on the value as a float
                    queryset = queryset \
                        .alias(numeric_value=Cast('value', output_field=FloatField())) \
                        .filter(type=MetaType.Numeric, **{f'numeric_value__{lookup}': value})
            else:
                errors['metadata_value'].append('This filter is not supported without metadata_name.')

        if errors:
            raise ValidationError(errors)

        return queryset.values('element_id')

    def get_filters(self):
        filters = {
            'corpus': self.selected_corpus
        }
        errors = {}

        if 'name' in self.request.query_params:
            filters['name__icontains'] = self.clean_params['name']

        if self.type_filter:
            filters['type'] = self.type_filter
        elif self.folder_filter is not None:
            filters['type__folder'] = self.folder_filter

        if 'worker_version' in self.clean_params:
            try:
                filters['worker_version_id'] = self.get_worker_version_filter()
            except ValidationError as e:
                errors['worker_version'] = e.detail

        try:
            metadata_queryset = self.get_metadata_queryset()
        except ValidationError as e:
            errors.update(e.detail)
        else:
            if metadata_queryset is not None:
                filters['id__in'] = metadata_queryset

        if errors:
            raise ValidationError(errors)

        return filters

    def get_classifications_filters(self):
        """
        Build Django ORM filters using Q expressions related to best classes.
        Supports 3 modes:
        - elements without any best classes
        - elements with any best classes
        - elements with a specific best class
        """
        class_filter = self.clean_params.get('best_class')
        if class_filter is None:
            return

        # Generic ORM query to find best classes:
        # - elements with a validated classification
        # - OR where high confidence is True and not rejected
        best_classifications = Q(classifications__state=ClassificationState.Validated) | (
            Q(classifications__high_confidence=True)
            & ~Q(classifications__state=ClassificationState.Rejected)
        )

        # List elements without any best classes, by inverting the query above
        if class_filter.lower() in ('false', '0'):
            return ~best_classifications

        try:
            # Filter on a specific class
            class_filter = uuid.UUID(class_filter)
            return best_classifications & Q(classifications__ml_class_id=class_filter)
        except (TypeError, ValueError):
            # By default, use all best classifications
            return best_classifications

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['corpus'] = self.selected_corpus
        if self.request:
            context['with_zone'] = self.request.query_params.get('with_zone', 'true').lower() not in ('false', '0')
            context['with_corpus'] = self.request.query_params.get('with_corpus', 'true').lower() not in ('false', '0')
        return context

    def get_prefetch(self):
        # Select the image and server in one separate query
        prefetch = {Prefetch('image', queryset=Image.objects.select_related('server')), 'type'}

        with_best_classes = self.clean_params.get('with_best_classes')
        if with_best_classes and with_best_classes.lower() not in ('false', '0'):
            prefetch.add(best_classifications_prefetch)

        return prefetch

    def get_order_direction(self):
        direction = self.clean_params.get('order_direction', 'asc').lower()
        if direction not in ('asc', 'desc'):
            raise ValidationError({'order_direction': ['Unknown sorting direction']})
        return '-' if direction == 'desc' else ''

    def get_order_by(self):
        sort_field = self.clean_params.get('order', 'name').lower()

        if sort_field not in ELEMENT_ORDERING_FIELDS:
            raise ValidationError({'order': ['Unknown sorting field']})

        direction = self.get_order_direction()
        # The sorting field is not unique; fallback on ordering by ID to ensure a consistent ordering
        return (f'{direction}{sort_field}', f'{direction}id')

    def filter_queryset(self, queryset):
        queryset = queryset \
            .filter(**self.get_filters()) \
            .prefetch_related(*self.get_prefetch()) \
            .order_by(*self.get_order_by())

        class_filters = self.get_classifications_filters()
        if class_filters is not None:
            # Use queryset.distinct() whenever best_class is defined
            queryset = queryset.filter(class_filters).distinct()

        with_has_children = self.clean_params.get('with_has_children')
        if with_has_children and with_has_children.lower() not in ('false', '0'):
            queryset = BulkMap(_fetch_has_children, queryset)

        return queryset

    def check_not_modified(self, queryset):
        """
        Returns True if HTTP 304 Not Modified should be returned. Requires a Django QuerySet instance or a BulkMap.
        """
        # Retrieve the inner QuerySet from the BulkMap if it is used
        if isinstance(queryset, BulkMap):
            queryset = queryset.iterable
        assert isinstance(queryset, QuerySet), 'A Django QuerySet is required to check for modified elements'

        modified_since_string = self.clean_params.get('If-Modified-Since')
        if not modified_since_string:
            return False

        # Parses RFC 2822 dates into tuples
        # Returns None if no date could be parsed or if the header was not defined
        time_tuple = email.utils.parsedate_tz(modified_since_string)
        if not time_tuple:
            raise ValidationError({'If-Modified-Since': ['Bad date format']})

        # Build a datetime from the tuple
        modified_since = datetime.fromtimestamp(email.utils.mktime_tz(time_tuple), timezone.utc)

        # Cause a HTTP 304 only if no element has been created or modified since then
        return not queryset.filter(Q(created__gt=modified_since) | Q(updated__gt=modified_since)).exists()

    def list(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())

        if self.check_not_modified(queryset):
            return Response(None, status=status.HTTP_304_NOT_MODIFIED)

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)

    def delete(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())
        delete_children = self.clean_params.get('delete_children', '').lower() not in ('false', '0')

        if not queryset.exists():
            raise NotFound

        element_trash(queryset, delete_children=delete_children, user_id=self.request.user.id)
        return Response(status=status.HTTP_204_NO_CONTENT)


@extend_schema(
    parameters=[
        OpenApiParameter(
            'top_level',
            description='Restrict to or exclude folder elements without parent elements (top-level elements).',
            type=bool,
            required=False,
        )
    ]
)
@extend_schema_view(
    get=extend_schema(operation_id='ListElements'),
    delete=extend_schema(operation_id='DestroyElements', description=(
        'Destroy elements in bulk.\n\n'
        'Requires an **admin** access to the corpus.'
    )),
)
class CorpusElements(ElementsListBase):
    """
    List elements in a corpus.

    Requires a **read** access to the corpus.
    """

    @property
    def is_top_level(self):
        if self.clean_params.get('top_level') is None:
            return None
        return self.clean_params['top_level'].lower() not in ('false', '0')

    def get_queryset(self):
        # Should not be possible due to the URL
        assert self.selected_corpus, 'Missing corpus ID'
        return self.selected_corpus.elements.all()

    def get_filters(self):
        filters = super().get_filters()

        if self.is_top_level is not None:
            # Filter elements with no parent or with any parent
            filters['paths__isnull'] = self.is_top_level

        return filters


@extend_schema(
    parameters=[
        OpenApiParameter(
            'recursive',
            type=bool,
            description='List recursively (parents, grandparents, etc.)',
            required=False,
        ),
    ]
)
@extend_schema_view(
    get=extend_schema(operation_id='ListElementParents'),
    delete=extend_schema(operation_id='DestroyElementParents', description=(
        'Delete parent elements in bulk.\n\n'
        "Requires an **admin** access to the element's corpus."
    )),
)
class ElementParents(ElementsListBase):
    """
    List all parents of an element.

    Requires a **read** access to the element's corpus.
    """

    @property
    def is_recursive(self):
        recursive_param = self.clean_params.get('recursive')
        return recursive_param is not None and recursive_param.lower() not in ('false', '0')

    @property
    def selected_corpus(self):
        return self.element.corpus

    def get_queryset(self):
        self.element = get_object_or_404(
            Element.objects.select_related('corpus'),
            id=self.kwargs['pk']
        )
        self.check_corpus_access(self.element.corpus)

        return Element.objects.get_ascending(self.kwargs['pk'], recursive=self.is_recursive)


@extend_schema(
    parameters=[
        OpenApiParameter(
            'recursive',
            type=bool,
            description='List recursively (children, grandchildren, etc.)',
            required=False,
        ),
    ]
)
@extend_schema_view(
    get=extend_schema(operation_id='ListElementChildren', parameters=[
        OpenApiParameter(
            'order',
            description='Sort elements by a specific field',
            enum=ELEMENT_ORDERING_FIELDS | {'position'},
            default='name',
            required=False,
        )
    ]),
    delete=extend_schema(operation_id='DestroyElementChildren', description=(
        'Delete child elements in bulk.\n\n'
        "Requires an **admin** access to the element's corpus."
    )),
)
class ElementChildren(ElementsListBase):
    """
    List all children of an element.

    Requires a **read** access to the element's corpus.
    """

    @property
    def selected_corpus(self):
        return self.element.corpus

    def get_filters(self):
        filters = super().get_filters()
        recursive_param = self.clean_params.get('recursive')

        if recursive_param is None or recursive_param.lower() in ('false', '0'):
            # Only list direct children, by listing all the direct paths from the current element
            # This is faster for direct children than :
            # - Element.objects.filter(paths__path__last=XXX)
            # - Element.objects.get_descending()
            # This code path is extremely used, as it's the basis for navigation amongst elements
            # Be extremely careful when changing this behaviour.
            paths = self.element.paths.values_list('path', flat=True)
            if not paths:
                # Support top level children
                paths = [[]]

            filters['paths__path__in'] = [
                path + [self.kwargs['pk']]
                for path in paths
            ]
        else:
            filters['paths__path__overlap'] = [self.kwargs['pk']]

        return filters

    def get_order_by(self):
        """
        Override the ElementsListBase ordering to default by path ordering.
        The only endpoint where a path ordering is made visible in the API
        is ListElementNeighbors, in a field named `position`.
        """
        if self.clean_params.get('order', 'position').lower() == 'position':
            direction = self.get_order_direction()
            return (f'{direction}paths__ordering', f'{direction}id')
        return super().get_order_by()

    def get_queryset(self):
        self.element = get_object_or_404(
            Element.objects.select_related('corpus'),
            id=self.kwargs['pk']
        )
        self.check_corpus_access(self.element.corpus)

        # Let `self.get_filters` handle filtering optimisations
        return Element.objects.all()


@extend_schema(tags=['elements'])
@extend_schema_view(
    get=extend_schema(description="Retrieve a single element's informations and metadata"),
    patch=extend_schema(description='Rename an element'),
    put=extend_schema(description="Edit an element's attributes. Requires a write access on the corpus."),
    delete=extend_schema(description='Delete an element. Requires either an admin access on the corpus, '
                                     'or a write access and to be the creator of this element.'),
)
class ElementRetrieve(ACLMixin, RetrieveUpdateDestroyAPIView):
    serializer_class = ElementSerializer
    permission_classes = (IsVerifiedOrReadOnly, )

    def get_object(self):
        # Prevent duplicating database request
        if not hasattr(self, 'element'):
            self.element = super().get_object()
        return self.element

    def get_queryset(self):
        corpora = Corpus.objects.readable(self.request.user)
        queryset = Element.objects.filter(corpus__in=corpora)
        if self.request and self.request.method == 'DELETE':
            # Only include corpus and creator for ACL check and ID for deletion
            return queryset.select_related('corpus').only('id', 'creator_id', 'corpus')

        return queryset \
            .select_related(
                'corpus',
                'type',
                'image__server',
                'creator',
            ) \
            .prefetch_related(Prefetch('classifications', queryset=classifications_queryset)) \
            .annotate(metadata_count=Count('metadatas'))

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)

        if request.method in permissions.SAFE_METHODS:
            return
        role = Role.Contributor
        if request.method == 'DELETE' and obj.creator_id != request.user.id:
            role = Role.Admin
        if not self.has_access(obj.corpus, role.value):
            access_repr = 'admin' if role == Role.Admin else 'write'
            raise PermissionDenied(detail=f'You do not have {access_repr} access to this element.')

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request and self.request.method != 'GET' and 'pk' in self.kwargs:
            context['element'] = self.get_object()
        return context

    def delete(self, request, *args, **kwargs):
        self.check_object_permissions(self.request, self.get_object())
        queryset = Element.objects.filter(id=self.kwargs['pk'])
        element_trash(queryset, user_id=self.request.user.id)
        return Response(status=status.HTTP_204_NO_CONTENT)


@extend_schema_view(
    get=extend_schema(
        operation_id='ListElementNeighbors',
        tags=['elements'],
        parameters=[
            OpenApiParameter(
                'n',
                type={'type': 'integer', 'minimum': 1, 'maximum': 10},
                description='Number of neighbors to retrieve around the element',
                required=False,
            )
        ],
    )
)
class ElementNeighbors(ACLMixin, ListAPIView):
    """
    List neighboring elements.

    Requires a **read** access to the element's corpus.
    """
    serializer_class = ElementNeighborsSerializer
    # For OpenAPI type discovery
    queryset = Element.objects.none()

    def get_queryset(self):
        n = self.request.query_params.get('n', 1)
        try:
            n = int(n)
        except (TypeError, ValueError):
            raise ValidationError({'n': 'Should be an integer between 1 and 10'})
        if not 1 <= n <= 10:
            raise ValidationError({'n': 'Should be an integer between 1 and 10'})

        element = get_object_or_404(
            Element.objects.select_related('corpus'),
            id=self.kwargs['pk']
        )

        # Check access permission
        if not self.has_access(element.corpus, Role.Guest.value):
            raise PermissionDenied(detail='You do not have a read access to this element.')

        return Element.objects.get_neighbors(element, n)


@extend_schema(tags=['elements'], request=None)
@extend_schema_view(
    post=extend_schema(operation_id='CreateElementParent', description='Link an element to a new parent'),
    delete=extend_schema(operation_id='DestroyElementParent', description='Delete the relation between an element and one of its parents'),
)
class ElementParent(CreateAPIView, DestroyAPIView):
    """
    Manage element relation with one of its parents
    """
    serializer_class = ElementParentSerializer
    permission_classes = (IsVerified, )

    def get_serializer_from_params(self, child=None, parent=None, **kwargs):
        data = {'child': child, 'parent': parent}
        kwargs['context'] = self.get_serializer_context()
        return ElementParentSerializer(data=data, **kwargs)

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer_from_params(**kwargs)
        serializer.is_valid(raise_exception=True)
        # Not necessary to serialize again
        serializer.create(serializer.validated_data)
        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)

    def destroy(self, request, *args, **kwargs):
        serializer = self.get_serializer_from_params(**kwargs)
        serializer.is_valid(raise_exception=True)
        serializer.delete(serializer.validated_data)
        return Response(status=status.HTTP_204_NO_CONTENT)


class RemoveSelectionAutoSchema(AutoSchema):
    """
    AutoSchema subclass that adds a requestBody for the RemoveSelection endpoint.
    This is not doable using only @extend_schema since a DELETE request shouldn't
    have a body according to OpenAPI rules.
    """

    def _get_request_body(self):
        if self.method != 'DELETE':
            return super()._get_request_body()

        # To bypass the no-body rule for DELETE methods, we force the DELETE method to POST.
        self.method = 'POST'
        request_body = super()._get_request_body()

        # The body isn't required but it will be ignored by APIStar, so we must always use body={} when calling RemoveSelection.
        request_body['required'] = False

        # Now that we have retrieved the requestBody schema, we can reset the method to its true nature.
        self.method = 'DELETE'
        return request_body


@extend_schema(tags=['elements'])
@extend_schema_view(
    get=extend_schema(
        operation_id='ListSelection',
        description='List all selected elements',
        parameters=[
            OpenApiParameter(
                'with_best_classes',
                description='Returns best classifications for each element. '
                            'If not set, elements best_classes field will always be null',
                type=bool,
                required=False,
            ),
        ],
    )
)
class ManageSelection(SelectionMixin, ListAPIView):
    """
    Manage all selected elements
    """
    schema = RemoveSelectionAutoSchema()
    serializer_class = ElementListSerializer
    permission_classes = (IsVerified, )

    def initial(self, request, *args, **kwargs):
        super().initial(request, *args, **kwargs)
        if not settings.ARKINDEX_FEATURES['selection']:
            raise ValidationError(['Selection is not available on this instance.'])

    def get_queryset(self):
        filtered_queryset = self.get_selection().select_related('type', 'corpus', 'image__server').order_by('corpus', 'type__slug', 'name', 'id')

        with_best_classes = self.request.query_params.get('with_best_classes')
        if with_best_classes and with_best_classes.lower() not in ('false', '0'):
            return filtered_queryset.prefetch_related(best_classifications_prefetch)

        return filtered_queryset

    # TODO: Use many=True on the response without causing Spectacular to see a paginated list
    @extend_schema(
        operation_id='AddSelection',
        description='Add specific elements',
        responses={201: ElementSlimSerializer},
        request=inline_serializer(
            name="AddSelectionBodySerializer",
            fields={'ids': serializers.ListField(child=serializers.UUIDField())}
        )
    )
    def post(self, request, *args, **kwargs):
        if 'ids' not in request.data:
            raise ValidationError({'ids': ['Element IDs are required.']})

        # Strange field validation done here,
        # as we have different serializers on each request method on the same endpoint.
        # Using serializer fields means we still preserve DRF's normal validation.
        field = serializers.ListField(child=serializers.UUIDField())
        element_ids = field.to_internal_value(request.data['ids'])

        elements = list(Element.objects.filter(
            id__in=element_ids,
            corpus__in=Corpus.objects.readable(request.user)
        ))

        if len(elements) != len(element_ids):
            raise ValidationError({'ids': [
                'Some element IDs do not exist or you do not have permission to access them.'
            ]})

        # Add new Selection objects in bulk, ignoring elements that were already selected
        Selection.objects.bulk_create(
            (
                Selection(user=self.request.user, element=element)
                for element in elements
            ),
            ignore_conflicts=True,
        )

        prefetch_related_objects(elements, 'corpus', 'image__server', 'type')
        return Response(
            status=status.HTTP_201_CREATED,
            data=ElementSlimSerializer(
                elements,
                context={'request': request},
                many=True
            ).data,
        )

    @extend_schema(
        operation_id='RemoveSelection',
        description='Remove a specific element or delete any selection',
        request=inline_serializer(
            name="RemoveSelectionBodySerializer",
            fields={'id': serializers.UUIDField(required=False)}
        )
    )
    def delete(self, request, *args, **kwargs):
        if request.data.get('id'):
            field = serializers.PrimaryKeyRelatedField(queryset=self.get_selection())
            element = field.to_internal_value(request.data['id'])
            request.user.selections.get(element=element).delete()
        else:
            request.user.selections.all().delete()

        return Response(status=status.HTTP_204_NO_CONTENT)


@extend_schema(tags=['corpora'])
@extend_schema_view(
    get=extend_schema(description='List corpora with their access rights'),
    post=extend_schema(description='Create a new corpus'),
)
class CorpusList(ListCreateAPIView):
    """
    List all available corpora and create new ones
    """
    serializer_class = CorpusSerializer
    pagination_class = None
    permission_classes = (IsVerifiedOrReadOnly, )

    def get_queryset(self):
        # We need to prefetch the acces rights on each corpora
        accessible_corpora = filter_rights(self.request.user, Corpus, Role.Guest.value)

        # Keep non duplicated corpora with the max access level
        corpora_level = defaultdict(int)
        for c in accessible_corpora:
            if c.max_level > corpora_level[c.id]:
                corpora_level[c.id] = c.max_level

        corpora = Corpus.objects \
            .filter(id__in=corpora_level) \
            .annotate(authorized_users=Count('memberships')) \
            .select_related('thumbnail__image__server') \
            .prefetch_related('types') \
            .order_by('name', 'id')

        for c in corpora:
            c.access_level = corpora_level[c.id]

        return corpora


@extend_schema(tags=['corpora'])
@extend_schema_view(
    get=extend_schema(description='Retrieve a single corpus'),
    put=extend_schema(description='Update a corpus'),
    patch=extend_schema(description='Partially update a corpus'),
    delete=extend_schema(
        description='Delete a corpus. Requires the "admin" right on the corpus. '
                    'This triggers an asynchronous deletion of a corpus and returns a '
                    'response immediately; the actual deletion will become visible once '
                    'the process is completed.'
    ),
)
class CorpusRetrieve(ACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Get, update and delete corpora
    """
    serializer_class = CorpusSerializer
    permission_classes = (IsVerifiedOrReadOnly, )

    def get_queryset(self):
        return Corpus.objects.readable(self.request.user) \
            .annotate(authorized_users=Count('memberships')) \
            .prefetch_related('types')

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)

        if self.request.method in permissions.SAFE_METHODS:
            return
        role = Role.Admin
        if not self.has_access(obj, role.value):
            access_repr = 'admin' if role == Role.Admin else 'write'
            raise PermissionDenied(detail=f'You do not have {access_repr} access to this corpus.')

    def perform_destroy(self, instance):
        corpus_delete(instance, user_id=self.request.user.id)


class TranscriptionsPagination(PageNumberPagination):
    page_size = 100


@extend_schema_view(
    get=extend_schema(
        operation_id='ListTranscriptions',
        tags=['transcriptions'],
        parameters=[
            OpenApiParameter(
                'recursive',
                type=bool,
                description='Recursively list transcriptions on sub-elements',
                required=False,
            ),
            OpenApiParameter(
                'worker_version',
                type=UUID,
                description='Filter transcriptions by worker version',
                required=False,
            ),
            OpenApiParameter(
                'element_type',
                description='Filter transcriptions by element type',
                required=False,
            )
        ]
    )
)
class ElementTranscriptions(ListAPIView):
    """
    List all transcriptions for an element, optionally filtered by type or worker version id.
    Recursive parameter allow listing transcriptions on sub-elements,
    otherwise element fields in the response will be set to null.
    """
    serializer_class = ElementTranscriptionSerializer
    pagination_class = TranscriptionsPagination
    queryset = Transcription.objects.none()

    @property
    def is_recursive(self):
        if not self.request:
            return
        recursive = self.request.query_params.get('recursive')
        return recursive is not None and recursive.lower() not in ('false', '0')

    def get_serializer_context(self):
        context = super().get_serializer_context()
        # Do serialize the element attached to each transcription in recursive mode only
        context['ignore_element'] = not self.is_recursive
        return context

    def get_queryset(self):
        element = get_object_or_404(Element.objects.filter(
            id=self.kwargs['pk'],
            corpus__in=Corpus.objects.readable(self.request.user)
        ))
        self.check_object_permissions(self.request, element)

        # ORDER BY casting IDs as char to avoid the PostgreSQL optimizer's inefficient scan
        # TODO: See if select_related is faster than a prefetch on this endpoint
        queryset = Transcription.objects \
            .prefetch_related('element__image__server', 'worker_version') \
            .annotate(char_id=Cast('id', output_field=CharField())) \
            .order_by('char_id')

        if self.is_recursive:
            queryset = queryset.filter(
                # Retrieve both element and sub-elements transcriptions
                element__in=[element.id, *Element.objects.get_descending(element.id).values_list('id')]
            ).prefetch_related('element__type')
        else:
            queryset = queryset.filter(element_id=element.id)

        return queryset

    def filter_queryset(self, queryset):
        # Filter by worker version
        worker_version = self.request.query_params.get('worker_version')
        if worker_version:
            try:
                queryset = queryset.filter(worker_version_id=worker_version)
            except DjangoValidationError as e:
                raise ValidationError({'worker_version': e.messages})

        # Filter by element_type
        element_type = self.request.query_params.get('element_type')
        if element_type:
            queryset = queryset.select_related('element__type').filter(element__type__slug=element_type)

        return queryset


@extend_schema_view(
    post=extend_schema(
        operation_id='CreateElement',
        tags=['elements'],
        parameters=[
            OpenApiParameter(
                'slim_output',
                type=bool,
                description="When enabled, the endpoint will only return the new Element's ID."
                            "This is especially useful for an automated tool, as it speeds up the element creation.",
                required=False,
            )
        ],
        # Return either a full element or the element ID depending on slim_output
        responses=PolymorphicProxySerializer(
            component_name='CreateElementResponse',
            resource_type_field_name='slim_output',
            serializers={
                True: inline_serializer('CreateElementSlimOutput', fields={'id': serializers.UUIDField()}),
                False: ElementSerializer,
            }
        )
    )
)
class ElementsCreate(CreateAPIView):
    """
    Create a new element
    """
    permission_classes = (IsVerified, )
    serializer_class = ElementCreateSerializer


@extend_schema_view(
    get=extend_schema(
        operation_id='ListElementMetaData',
        description='List all metadata linked to an element.',
        tags=['elements'],
    ),
    post=extend_schema(
        operation_id='CreateMetaData',
        tags=['elements'],
    )
)
class ElementMetadata(ListCreateAPIView):
    """
    Create a metadata on an existing element.

    The `date` type allows dates to be parsed and indexed for search APIs. The supported date formats are as follows:

    * `YYYY`: `1995` (recommended)
    * `YYYY-MM`: `1995-02` (recommended)
    * `YYYY-MM-DD`: `1995-02-16` (recommended)
    * `/YYYY(.*)month/`: `1995, february`, with months in French or English (case-insensitive)
    * `YYYY-YYYY`: `1995-1996` (yields a date interval)

    The `reference` type is also indexed in search APIs and allows searching by
    known identifiers or references, or any arbitrary string.
    """
    permission_classes = (IsVerifiedOrReadOnly, )
    pagination_class = None
    serializer_class = MetaDataUpdateSerializer

    def get_queryset(self):
        if self.request and self.request.method == 'GET':
            element = get_object_or_404(Element, id=self.kwargs['pk'], corpus__in=Corpus.objects.readable(self.request.user))
            return element.metadatas.all().select_related('entity')

        # We need to use the default database to avoid stale read
        # when a metadata is created immediately after an element creation.
        # Note that this returns a queryset of elements and not metadata!
        return Element.objects.using('default').filter(corpus__in=Corpus.objects.writable(self.request.user))

    def get_serializer_context(self):
        context = super().get_serializer_context()
        # Ignore this step when generating the schema with OpenAPI
        if not self.request:
            return context
        if self.request.method != 'GET':
            context['element'] = self.get_object()
        return context

    def perform_create(self, serializer):
        instance = serializer.save()
        if self.request.user.is_admin or self.request.user.is_internal:
            AllowedMetaData.objects.get_or_create(name=instance.name, type=instance.type, corpus=instance.element.corpus)


@extend_schema_view(
    post=extend_schema(
        operation_id='CreateMetaDataBulk',
        tags=['elements'],
    )
)
class ElementMetadataBulk(CreateAPIView):
    """
    Create multiple metadata on an existing element.
    """
    permission_classes = (IsVerified, )
    serializer_class = MetaDataBulkSerializer

    def get_queryset(self):
        # We need to use the default database to avoid stale read
        # when metadata are created immediately after an element creation
        return Element.objects.using('default').filter(corpus__in=Corpus.objects.writable(self.request.user))

    def get_serializer_context(self):
        context = super().get_serializer_context()
        # Ignore this step when generating the schema with OpenAPI
        if not self.request:
            return context
        context['element'] = self.get_object()
        return context

    def perform_create(self, serializer):
        instances = serializer.save()
        if self.request.user.is_admin or self.request.user.is_internal:
            AllowedMetaData.objects.bulk_create(
                [
                    AllowedMetaData(name=instance.name, type=instance.type, corpus=instance.element.corpus)
                    for instance in instances['metadata_list']
                ],
                ignore_conflicts=True
            )


@extend_schema(tags=['elements'])
@extend_schema_view(
    get=extend_schema(operation_id='RetrieveMetaData', description='Retrieve an existing metadata'),
    patch=extend_schema(operation_id='PartialUpdateMetaData', description='Partially update an existing metadata'),
    put=extend_schema(operation_id='UpdateMetaData', description='Update an existing metadata'),
    delete=extend_schema(operation_id='DestroyMetaData', description='Delete an existing metadata')
)
class MetadataEdit(ACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Edit an existing element medadata
    """
    permission_classes = (IsVerified, )
    serializer_class = MetaDataUpdateSerializer

    def get_queryset(self):
        # Filter readable metadata in order to check object permissions
        return MetaData.objects.select_related('element__corpus').filter(element__corpus__in=Corpus.objects.readable(self.request.user))

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)
        if not self.has_access(obj.element.corpus, Role.Contributor.value):
            self.permission_denied(request, message='You do not have write access to this corpus.')

    def perform_destroy(self, instance):
        serializer = self.get_serializer(instance)
        serializer.delete(instance)


class AllowedMetaDataPagination(PageNumberPagination):
    page_size = 200


@extend_schema(tags=['corpora'])
@extend_schema_view(
    get=extend_schema(operation_id='ListCorpusAllowedMetaDatas'),
    post=extend_schema(
        operation_id='CreateAllowedMetaData',
        description='Create an allowed meta data in a corpus',
    )
)
class CorpusAllowedMetaData(CorpusACLMixin, ListCreateAPIView):
    """
    List (type, name) couples allowed for corpus metadata
    """
    serializer_class = CorpusAllowedMetaDataSerializer
    pagination_class = AllowedMetaDataPagination
    permission_classes = (IsVerifiedOrReadOnly, )
    queryset = AllowedMetaData.objects.none()

    @cached_property
    def corpus(self):
        role = Role.Guest
        if self.request.method.lower() == 'post':
            role = Role.Admin
        return self.get_corpus(self.kwargs['pk'], role=role)

    def get_queryset(self):
        return AllowedMetaData.objects.filter(corpus=self.corpus)

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['corpus_id'] = self.corpus.id
        return context


@extend_schema(tags=['corpora'])
@extend_schema_view(
    get=extend_schema(operation_id='RetrieveAllowedMetaData', description='Retrieve an allowed metadata.'),
    patch=extend_schema(operation_id='PartialUpdateAllowedMetaData', description='Edit an allowed metadata.'),
    put=extend_schema(operation_id='UpdateAllowedMetaData', description='Edit an allowed metadata.'),
    delete=extend_schema(operation_id='DestroyAllowedMetaData', description='Delete an existing allowed metadata.')
)
class AllowedMetaDataEdit(CorpusACLMixin, RetrieveUpdateDestroyAPIView):
    permission_classes = (IsVerified, )
    serializer_class = CorpusAllowedMetaDataSerializer
    queryset = AllowedMetaData.objects.none()

    @cached_property
    def corpus(self):
        role = Role.Guest
        if self.request.method.lower() not in permissions.SAFE_METHODS:
            role = Role.Admin
        return self.get_corpus(self.kwargs['corpus'], role=role)

    def get_queryset(self):
        return AllowedMetaData.objects.filter(corpus=self.corpus)

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['corpus_id'] = self.corpus.id
        return context


@extend_schema(tags=['elements'])
class ElementTypeCreate(CreateAPIView):
    """
    Create a new element type for a specific corpus
    """
    serializer_class = ElementTypeSerializer
    permission_classes = (IsVerifiedOrReadOnly, )


@extend_schema_view(
    put=extend_schema(operation_id='UpdateElementType', tags=['elements']),
    patch=extend_schema(operation_id='PartialUpdateElementType', tags=['elements']),
)
class ElementTypeUpdate(UpdateAPIView):
    """
    Update an existing element type
    """
    serializer_class = ElementTypeLightSerializer
    permission_classes = (IsVerifiedOrReadOnly, )

    def get_queryset(self):
        return ElementType.objects.filter(corpus__in=Corpus.objects.readable(self.request.user))


@extend_schema_view(
    post=extend_schema(
        operation_id='CreateElements',
        tags=['elements'],
        responses=inline_serializer(
            'ElementBulkCreateResponse',
            fields={'id': serializers.UUIDField()},
            many=True,
        ),
    )
)
class ElementBulkCreate(CreateAPIView):
    """
    Create multiple child elements at once on a single parent
    """
    serializer_class = ElementBulkSerializer
    permission_classes = (IsVerified, )
    pagination_class = None

    def get_object(self):
        if not hasattr(self, 'element'):
            self.element = super().get_object()
            if not self.element.image_id:
                raise ValidationError({'element': ['Cannot create child elements on an element without a zone.']})
        return self.element

    def get_queryset(self):
        if not self.request:
            return Element.objects.none()
        return Element \
            .objects \
            .using('default') \
            .filter(corpus__in=Corpus.objects.writable(self.request.user)) \
            .only('id', 'corpus_id', 'image_id')

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request:
            context['element'] = self.get_object()
        return context

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        data = self.perform_create(serializer)
        headers = self.get_success_headers({})
        return Response(data, status=status.HTTP_201_CREATED, headers=headers)

    @transaction.atomic
    def perform_create(self, serializer):
        worker_version = serializer.validated_data.get('worker_version')
        elements = serializer.validated_data['elements']

        # ElementPath.get_next_order, but for multiple types
        next_orderings = defaultdict(
            int,
            ElementPath
            .objects
            .using('default')
            .filter(
                # Using __overlap before __last makes Postgres use the GIN index instead of scanning the whole table
                path__overlap=[self.element.id],
                path__last=self.element.id,
                element__type_id__in=set(element_data['type'] for element_data in elements)
            )
            .values('element__type_id')
            .annotate(next_ordering=Max('ordering') + 1)
            .values_list('element__type_id', 'next_ordering')
        )
        # Prepare elements
        for element_data in elements:
            # Keep ordering in the element payload to optimize element path creation later
            element_data['ordering'] = next_orderings[element_data['type']]
            next_orderings[element_data['type']] += 1

            element_data['element'] = Element(
                id=uuid.uuid4(),
                corpus_id=self.element.corpus_id,
                type_id=element_data['type'],
                name=element_data['name'],
                worker_version=worker_version,
                image_id=self.element.image_id,
                polygon=element_data['polygon'],
            )

        Element.objects.bulk_create([
            element_data['element'] for element_data in elements
        ])

        # Support top level elements, by adding an empty initial path to trigger loops below
        parent_paths = list(self.element.paths.all().values_list('path', flat=True))
        if not parent_paths:
            parent_paths = [[]]

        ElementPath.objects.bulk_create([
            ElementPath(
                path=parent_path + [self.element.id],
                element_id=element_data['element'].id,
                ordering=element_data['ordering'],
            )
            for parent_path in parent_paths
            for element_data in elements
        ])

        return [{'id': element_data['element'].id} for element_data in elements]


@extend_schema_view(
    delete=extend_schema(
        operation_id='DestroyCorpusSelection',
        tags=['elements'],
    )
)
class CorpusSelectionDestroy(CorpusACLMixin, SelectionMixin, DestroyAPIView):
    """
    Delete all selected elements on a corpus
    """
    serializer_class = CorpusSerializer
    permission_classes = (IsVerified, )

    def delete(self, request, *args, **kwargs):
        corpus = self.get_corpus(self.kwargs['pk'], role=Role.Admin)
        selected_elements = self.get_selection(corpus_id=corpus.id)
        if not selected_elements.exists():
            raise NotFound

        for batch in range(0, selected_elements.count(), 50):
            queryset = Element.objects.filter(id__in=list(selected_elements[batch:batch + 50].values_list('id', flat=True)))
            element_trash(queryset, user_id=self.request.user.id)

        return Response(status=status.HTTP_204_NO_CONTENT)


@extend_schema_view(
    delete=extend_schema(
        operation_id='DestroyWorkerResults',
        parameters=[
            OpenApiParameter(
                'element_id',
                description='Only delete Worker Results under this parent element',
            )
        ],
        tags=['elements'],
    )
)
class WorkerResultsDestroy(CorpusACLMixin, DestroyAPIView):
    """
    Delete all Worker Results for a WorkerVersion on a Corpus or under a specified parent element (parent element included)
    """
    permission_classes = (IsVerified, )

    # https://github.com/tfranzel/drf-spectacular/issues/308
    @extend_schema(responses={204: None})
    def delete(self, request, *args, **kwargs):
        corpus = self.get_corpus(self.kwargs['corpus'], role=Role.Admin)
        try:
            version = WorkerVersion.objects.select_related('worker', 'revision').get(id=self.kwargs['version'])
        except WorkerVersion.DoesNotExist:
            raise ValidationError({'worker_version_id': ['This version does not exist.']})

        element_id = None
        if 'element_id' in self.request.query_params:
            try:
                element_id = UUID(self.request.query_params['element_id'])
            except (TypeError, ValueError):
                raise ValidationError({'element_id': ['Invalid UUID.']})

            if not Element.objects.filter(id=element_id).exists():
                raise ValidationError({'element_id': ['This element does not exist.']})

        worker_results_delete(
            corpus_id=corpus.id,
            version=version,
            parent_id=element_id,
            user_id=self.request.user.id
        )

        return Response(status=status.HTTP_204_NO_CONTENT)


@extend_schema_view(
    post=extend_schema(operation_id='MoveElement', tags=['elements']),
)
class ElementMove(CreateAPIView):
    """
    Move an element to a new destination folder
    """
    serializer_class = ElementDestinationSerializer
    permission_classes = (IsVerified, )

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        source = serializer.validated_data['source']
        destination = serializer.validated_data['destination']
        serializer.perform_create_checks(source, destination)

        move_element(source=source, destination=destination, user_id=self.request.user.id)

        return Response(serializer.data, status=status.HTTP_200_OK)
