import bleach
from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from arkindex.documents.dates import DateType
from arkindex.documents.models import AllowedMetaData, Corpus, Element, ElementType, MetaData, MetaType
from arkindex.images.serializers import ZoneLightSerializer
from arkindex.project.serializer_fields import EnumField


class InterpretedDateSerializer(serializers.Serializer):
    """
    Serialize a list of interpreted dates linked to a metadata
    """
    type = EnumField(DateType)
    year = serializers.IntegerField(min_value=0)
    month = serializers.IntegerField(min_value=1, max_value=12, allow_null=True)
    day = serializers.IntegerField(min_value=1, max_value=31, allow_null=True)

    class Meta:
        fields = (
            'year',
            'month',
            'day',
            'type',
        )


class ElementLightSerializer(serializers.ModelSerializer):
    """
    Serialises a Element
    """
    type = serializers.CharField(source='type.slug')

    class Meta:
        model = Element
        fields = (
            'id',
            'type',
            'name',
        )


class ElementZoneSerializer(ElementLightSerializer):
    """
    Lightly serialises an element with its type and zone
    """
    zone = serializers.SerializerMethodField(read_only=True)

    @extend_schema_field(ZoneLightSerializer(read_only=True))
    def get_zone(self, element):
        """
        Elements used to have zones, which held images and polygons.
        Those attributes are now directly on elements. While we could just use
        ZoneLightSerializer(source='*') in the attributes directly, this would cause
        the ZoneLightSerializer to always be present. Instead of having zone=None,
        we would see zone={'image': None, 'polygon': None, 'url': None}.
        This method ensures the zone attribute is None.
        """
        if not element.image_id or not element.polygon:
            return
        return ZoneLightSerializer(element).data

    class Meta(ElementLightSerializer.Meta):
        model = Element
        fields = ElementLightSerializer.Meta.fields + ('zone', )


class ElementTypeLightSerializer(serializers.ModelSerializer):

    class Meta:
        model = ElementType
        fields = (
            'id',
            'slug',
            'display_name',
            'folder',
        )

    def validate(self, data):
        slug = data.get('slug') if 'slug' in data else self.instance.slug
        corpus = data.get('corpus') if 'corpus' in data else self.instance.corpus
        type = corpus.types.filter(slug=slug)
        if type.exists() and (not self.instance or type.first().id != self.instance.id):
            raise ValidationError({'slug': ['Slug must be unique']})
        return data


class CorpusLightSerializer(serializers.ModelSerializer):
    """
    Serializes a Corpus
    """

    class Meta:
        model = Corpus
        fields = (
            'id',
            'name',
            'public',
        )


class MetaDataLightSerializer(serializers.ModelSerializer):
    """
    Serialize a metadata without its entity
    """
    type = EnumField(MetaType)
    dates = InterpretedDateSerializer(many=True, source='get_dates', read_only=True)

    def to_representation(self, instance):
        # Clean metadata value
        tags = []
        if instance.type == MetaType.Markdown:
            tags = [
                'a', 'b', 'blockquote', 'body', 'br', 'div',
                'em', 'h1', 'h2', 'h3', 'html', 'i', 'iframe',
                'img', 'li', 'marquee', 'ol', 'p', 'sup',
                'table', 'tbody', 'td', 'th', 'thead', 'tr',
                'ul'
            ]
        instance.value = bleach.clean(instance.value, tags=tags)
        return super().to_representation(instance)

    def check_allowed(self, corpus, type=None, name=None, instance=None, **kwargs):
        """
        Assert metadata type and value are allowed in this corpus
        """
        message = f'Metadata with this type ({type}) and name ({name}) are not allowed for this corpus'
        if instance:
            if type is None:
                type = instance.type
            if name is None:
                name = instance.name
            message = f'Edition of metadata with this type ({type}) and name ({name}) is not allowed for this corpus'

        # Ensure there always is a type and a name. Otherwise, an endpoint might make the user believe an allowed
        # metadata is missing when there is just a bug in our code, because Django does not mind filtering by None.
        assert type is not None and name is not None, 'Both a type and name are required to check for allowed metadata'

        if not AllowedMetaData.objects.filter(corpus=corpus, type=type, name=name):
            raise ValidationError({
                '__all__': [message]
            })

    def create(self, validated_data):
        user = self.context['request'].user
        element = self.context['element']

        if not (user.is_admin or user.is_internal):
            self.check_allowed(corpus=element.corpus, **validated_data)

        try:
            existing_metadata = element.metadatas.using('default').filter(
                name=validated_data['name'],
                type=validated_data['type'],
                value=validated_data['value'],
            ).get()
        except MetaData.DoesNotExist:
            # No existing metadata found: carry on!
            pass
        except MetaData.MultipleObjectsReturned:
            # In theory, we should have a unique constraint on element_id/name/type/value, but that isn't possible
            # because the values can get very long and exceed the size of a B-Tree index tuple (8/3 kibibytes).
            # The APIs have unicity checks in place so this should never happen, but should is not enough.
            # If there are any duplicate metadatas, we just let the exception escalate to Sentry so that we can troubleshoot.
            raise
        else:
            # Send the existing ID in an HTTP 400 error
            raise ValidationError({
                'detail': ['This metadata already exists.'],
                'id': str(existing_metadata.id),
            })

        return element.metadatas.create(**validated_data)

    def update(self, instance, validated_data):
        user = self.context['request'].user
        element = instance.element
        if not (user.is_admin or user.is_internal):
            # Assert actual instance is part of AllowedMetaData
            self.check_allowed(corpus=element.corpus, instance=instance)
            # Assert the new values is part of AllowedMetaData
            # Still provide the instance so that the check can rely on the original values for partial updates
            self.check_allowed(corpus=element.corpus, instance=instance, **validated_data)

        return super().update(instance, validated_data)

    def delete(self, instance):
        user = self.context['request'].user
        if not (user.is_admin or user.is_internal):
            self.check_allowed(corpus=instance.element.corpus, instance=instance)

        instance.delete()

    class Meta:
        model = MetaData
        read_only_fields = ('id', )
        fields = (
            'id',
            'type',
            'name',
            'value',
            'dates',
        )


class CorpusAllowedMetaDataSerializer(serializers.ModelSerializer):
    """
    Serialize allowed (type, name) couples for metadata creation in a corpus
    """
    type = EnumField(MetaType)

    class Meta:
        model = AllowedMetaData
        fields = ('id', 'type', 'name', )
        read_only_fields = ('id', )

    def validate(self, data):
        data = super().validate(data)
        name, type = None, None
        if self.instance:
            name = self.instance.name
            type = self.instance.type
        name = data.get('name', name)
        type = data.get('type', type)
        if AllowedMetaData.objects.filter(corpus_id=self.context['corpus_id'], name=name, type=type).exists():
            raise serializers.ValidationError({
                'detail': ['An AllowedMetaData with this type and name already exists in this corpus.'],
            })
        data['corpus_id'] = self.context['corpus_id']
        return data
